# Copyright 2017,2021 IBM Corp. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""
Unit tests for _utils module.
"""


import os
import sys
from datetime import datetime, timezone, tzinfo, MAXYEAR
try:
    from zoneinfo import ZoneInfo
except ImportError:
    ZoneInfo = None
import time
import re
import pytest
from zhmcclient_mock import FakedSession
from zhmcclient import Client, FilterConversionError
from zhmcclient._utils import datetime_from_timestamp, \
    timestamp_from_datetime, datetime_to_isoformat, datetime_from_isoformat, \
    matches_filters, divide_filter_args, tzlocal

# datetime.fromisoformnat() supports 'hhmm' without colon as timezone offset
ISOFORMAT_SUPPORTS_HHMM = tuple(map(int, sys.version_info[:2])) >= (3, 11)

# The Unix epoch
EPOCH_DT = datetime(1970, 1, 1, 0, 0, 0, 0, timezone.utc)

# HMC timestamp numbers (milliseconds since the Unix epoch)
TS_2000_02_29 = 951782400000  # last day in february in leap year
TS_2001_02_28 = 983318400000  # last day in february in non-leap year
TS_2017_08_15 = 1502755200000  # some day in summer
TS_2038_LIMIT = 2147483647000  # 2038-01-19 03:14:07 UTC (Year 2038 Problem)
TS_3000_12_31 = 32535129600000
TS_3001_LIMIT = 32535244799000  # 3001-01-01 07:59:59 UTC (Visual C++ limit)
TS_MAX = 253402300799999  # 9999-12-31 23:59:59.999 UTC

# Some durations in milliseconds
SEC_MS = 1000
MIN_MS = SEC_MS * 60
HOUR_MS = MIN_MS * 60
DAY_MS = HOUR_MS * 24

# Test cases for datetime / timestamp conversion functions
DATETIME_TIMESTAMP_TESTCASES = [
    # (datetime_tuple(y,m,d,h,m,s,us), timestamp)
    ((1970, 1, 1, 0, 0, 0, 0), 0),
    ((1970, 1, 1, 0, 0, 0, 123000), 123),
    ((1970, 1, 1, 0, 2, 3, 456000), 123456),
    ((1970, 1, 2, 10, 17, 36, 789000), 123456789),
    ((1973, 11, 29, 21, 33, 9, 123000), 123456789123),
    ((2000, 2, 29, 0, 0, 0, 0), TS_2000_02_29),
    ((2000, 3, 1, 0, 0, 0, 0), TS_2000_02_29 + DAY_MS),
    ((2001, 2, 28, 0, 0, 0, 0), TS_2001_02_28),
    ((2001, 3, 1, 0, 0, 0, 0), TS_2001_02_28 + DAY_MS),
    ((2017, 8, 15, 0, 0, 0, 0), TS_2017_08_15),
    ((2038, 1, 19, 3, 14, 7, 0), TS_2038_LIMIT),
    ((2038, 1, 19, 3, 14, 7, 1000), TS_2038_LIMIT + 1),
    ((3000, 12, 31, 23, 59, 59, 999000), TS_3000_12_31 + DAY_MS - 1),
    ((3001, 1, 1, 7, 59, 59, 0), TS_3001_LIMIT),
    ((3001, 1, 1, 8, 0, 0, 0), TS_3001_LIMIT + 1000),
    ((MAXYEAR - 1, 12, 31, 0, 0, 0, 0), TS_MAX + 1 - DAY_MS - 365 * DAY_MS),
    ((MAXYEAR, 12, 30, 0, 0, 0, 0), TS_MAX + 1 - 2 * DAY_MS),
    ((MAXYEAR, 12, 30, 23, 59, 59, 0),
     TS_MAX + 1 - 2 * DAY_MS + 23 * HOUR_MS + 59 * MIN_MS + 59 * SEC_MS),
    ((MAXYEAR, 12, 31, 0, 0, 0, 0), TS_MAX + 1 - 1 * DAY_MS),
]


def find_max_value(test_func, initial_value):
    """
    Starting from an initial number (integer or float), find the maximum value
    for which the test function does not yet fail, and return that maximum
    value.
    """
    assert isinstance(initial_value, int) and initial_value > 0

    fails = FailsArray(test_func)
    value = initial_value

    # Advance the value exponentially beyond the max value
    while fails[value] == 0:
        value *= 2

    # Search for the exact max value in the previous range. We search for the
    # boundary where the fails array goes from 0 to 1.
    boundary = 0.5
    value = binary_search(fails, boundary, value // 2, value)
    max_value = value - 1

    # Verify that we found exactly the maximum:
    assert fails[max_value] == 0 and fails[max_value + 1] == 1, (
        f"max_value={max_value}, fails[+-2]: {fails[max_value - 2]}, "
        f"{fails[max_value - 1]}, {fails[max_value]}, "
        f"{fails[max_value + 1]}, {fails[max_value + 2]}")

    return max_value


class FailsArray:
    # pylint: disable=too-few-public-methods
    """
    An array that when accessed at an index returns 1 if the array value at
    that index causes an exception to be raised when passed to a test function,
    and 0 otherwise.
    """

    def __init__(self, test_func):
        """
        test_func is a test function that gets one integer argument and that
        raises an exception if the argument is out of range.
        """
        self.test_func = test_func

    def __getitem__(self, test_value):
        """
        Array access function, returning 1 if the given test value caused the
        test function to fail, and 0 otherwise.
        """
        try:
            self.test_func(test_value)
        except Exception:  # pylint: disable=broad-except
            return 1
        return 0


def binary_search(haystack, needle, lo, hi):
    """
    Binary search: Return index of value needle (or the next one if no exact
    match) in array haystack, within index range lo to hi.
    """
    while lo < hi:
        mid = (lo + hi) // 2
        if haystack[mid] > needle:
            hi = mid
        elif haystack[mid] < needle:
            lo = mid + 1
        else:
            return mid
    return hi


# TODO: Add testcases for repr_text() (currently indirectly tested)
# TODO: Add testcases for repr_list() (currently indirectly tested)
# TODO: Add testcases for repr_dict() (currently indirectly tested)
# TODO: Add testcases for repr_timestamp() (currently indirectly tested)
# TODO: Add testcases for repr_manager() (currently indirectly tested)


# Some tests for Python date & time related functions that we use.

def test_gmtime_epoch():
    """Test that time.gmtime() is based upon the UNIX epoch."""
    epoch_st = time.struct_time([1970, 1, 1, 0, 0, 0, 3, 1, 0])
    st = time.gmtime(0)
    assert st == epoch_st


def x_test_print_gmtime_max():
    """Print the maximum for time.gmtime()."""
    max_ts = find_max_value(time.gmtime, 1)
    max_st = time.gmtime(max_ts)
    print("\nMax Unix timestamp value for Python time.gmtime(): "
          f"{max_ts} ({max_st!r})")
    sys.stdout.flush()


def x_test_print_fromtimestamp_max():
    """Print the maximum for datetime.fromtimestamp(utc)."""

    def datetime_fromtimestamp_utc(ts):
        return datetime.fromtimestamp(ts, timezone.utc)

    max_ts = find_max_value(datetime_fromtimestamp_utc, 1)
    max_dt = datetime_fromtimestamp_utc(max_ts)
    print("\nMax Unix timestamp value for Python "
          f"datetime.fromtimestamp(utc): {max_ts} ({max_dt!r})")
    sys.stdout.flush()


def x_test_print_datetime_max():
    """Print datetime.max."""
    print(f"\nMax value for Python datetime (datetime.max): {datetime.max!r}")
    sys.stdout.flush()


# All tests for the datetime_from_timestamp() function.

@pytest.mark.parametrize(
    "tz_name", [None, 'UTC', 'US/Eastern', 'Europe/Berlin']
)
@pytest.mark.parametrize(
    "datetime_tuple, timestamp", DATETIME_TIMESTAMP_TESTCASES
)
def test_success_datetime_from_timestamp(
        datetime_tuple, timestamp, tz_name):
    """Test successful calls to datetime_from_timestamp()."""

    if os.name == 'nt' and timestamp > TS_3001_LIMIT:
        # Skip this test case, due to the lower limit on Windows
        return

    # Expected result, as timezone-unaware (but implied UTC)
    exp_dt_unaware = datetime(*datetime_tuple)

    # Expected result, as timezone-aware
    exp_dt = exp_dt_unaware.replace(tzinfo=timezone.utc)

    if tz_name is None:

        # Execute the code to be tested
        act_dt = datetime_from_timestamp(timestamp)

    else:
        if ZoneInfo is None:
            pytest.skip("ZoneInfo is not supported on this Python version")

        tz = ZoneInfo(tz_name)

        # Execute the code to be tested
        act_dt = datetime_from_timestamp(timestamp, tz)

    # Verify the result.
    # Note: timezone-aware datetime objects compare equal according to
    # their effective point in time (e.g. as if normalized to UTC).
    assert act_dt == exp_dt


@pytest.mark.parametrize(
    "timestamp, exc_type", [
        (None, ValueError),
        (-1, ValueError),
        (TS_MAX + 1, ValueError),
    ]
)
def test_error_datetime_from_timestamp(timestamp, exc_type):
    """Test failing calls to datetime_from_timestamp()."""

    with pytest.raises(Exception) as exc_info:

        # Execute the code to be tested
        datetime_from_timestamp(timestamp)

    # Verify the result
    assert isinstance(exc_info.value, exc_type)


def x_test_print_max_datetime_from_timestamp():
    """Print the maximum for datetime_from_timestamp()."""
    max_ts = find_max_value(datetime_from_timestamp, 1)
    max_dt = datetime_from_timestamp(max_ts)
    print("\nMax HMC timestamp value for zhmcclient."
          f"datetime_from_timestamp(): {max_ts} ({max_dt!r})")
    sys.stdout.flush()


# All tests for the timestamp_from_datetime() function.

@pytest.mark.parametrize(
    "tz_name", [None, 'UTC', 'US/Eastern', 'Europe/Berlin']
)
@pytest.mark.parametrize(
    "datetime_tuple, timestamp", DATETIME_TIMESTAMP_TESTCASES
)
def test_success(datetime_tuple, timestamp, tz_name):
    """Test successful calls to timestamp_from_datetime()."""

    # Create a timezone-naive datetime object
    dt = datetime(*datetime_tuple)
    offset = 0  # because of default UTC

    if tz_name is not None:
        # Make the datetime object timezone-aware
        if ZoneInfo is None:
            pytest.skip("ZoneInfo is not supported on this Python version")
        tz = ZoneInfo(tz_name)
        offset = tz.utcoffset(dt).total_seconds()
        dt = dt.replace(tzinfo=tz)

    exp_ts = timestamp - offset * 1000

    # Execute the code to be tested
    ts = timestamp_from_datetime(dt)

    # Verify the result
    assert ts == exp_ts


@pytest.mark.parametrize(
    "datetime_, exc_type", [
        (None, ValueError),
    ]
)
def test_error(datetime_, exc_type):
    """Test failing calls to timestamp_from_datetime()."""

    with pytest.raises(Exception) as exc_info:

        # Execute the code to be tested
        timestamp_from_datetime(datetime_)

    # Verify the result
    assert isinstance(exc_info.value, exc_type)


def test_datetime_max():
    """Test timestamp_from_datetime() with datetime.max."""

    # The test is that it does not raise an exception:
    timestamp_from_datetime(datetime.max)


# Test cases for datetime_to_isoformat()
TESTCASES_DATETIME_TO_ISOFORMAT = [
    # dt, exp_dt_str
    (
        datetime(2017, 9, 5, 12, 13, 10, 0),
        '2017-09-05 12:13:10'
    ),
    (
        datetime(2017, 9, 5, 12, 13, 10, 123456),
        '2017-09-05 12:13:10.123456'
    ),
    (
        datetime(2017, 9, 5, 12, 13, 10, 0, timezone.utc),
        '2017-09-05 12:13:10+00:00'
    ),
]
if ZoneInfo is not None:
    TESTCASES_DATETIME_TO_ISOFORMAT.extend(
        [
            (
                datetime(2017, 9, 5, 12, 13, 10, 0, ZoneInfo('Europe/Berlin')),
                '2017-09-05 12:13:10+02:00'  # DST
            ),
            (
                datetime(2017, 9, 5, 12, 13, 10, 0,
                         ZoneInfo('America/New_York')),
                '2017-09-05 12:13:10-04:00'  # DST
            ),
        ]
    )


@pytest.mark.parametrize(
    "dt, exp_dt_str",
    TESTCASES_DATETIME_TO_ISOFORMAT)
def test_datetime_to_isoformat(dt, exp_dt_str):
    """
    Test function for datetime_to_isoformat().
    """

    # The function to be tested
    dt_str = datetime_to_isoformat(dt)

    assert dt_str == exp_dt_str


# Test cases for datetime_from_isoformat()
TESTCASES_DATETIME_FROM_ISOFORMAT = [
    # dt_str, exp_dt
    (
        '2017-09-05 12:13:10',
        datetime(2017, 9, 5, 12, 13, 10, 0)  # timezone-naive
    ),
    (
        '2017-09-05T12:13:10',
        datetime(2017, 9, 5, 12, 13, 10, 0)  # timezone-naive
    ),
    (
        '2017-09-05 12:13:10.123456',
        datetime(2017, 9, 5, 12, 13, 10, 123456)  # timezone-naive
    ),
    (
        '2017-09-05 12:13:10+00:00',
        datetime(2017, 9, 5, 12, 13, 10, 0, timezone.utc)
    ),
]
if ISOFORMAT_SUPPORTS_HHMM:
    TESTCASES_DATETIME_FROM_ISOFORMAT.extend(
        [
            (
                '2017-09-05 12:13:10+0000',
                datetime(2017, 9, 5, 12, 13, 10, 0, timezone.utc)
            ),
        ]
    )
if ZoneInfo is not None:
    TESTCASES_DATETIME_FROM_ISOFORMAT.extend(
        [
            (
                '2017-09-05 12:13:10+02:00',  # DST
                datetime(2017, 9, 5, 12, 13, 10, 0, ZoneInfo('Europe/Berlin'))
            ),
            (
                '2017-09-05 12:13:10-04:00',  # DST
                datetime(2017, 9, 5, 12, 13, 10, 0,
                         ZoneInfo('America/New_York'))
            ),
        ]
    )
    if ISOFORMAT_SUPPORTS_HHMM:
        TESTCASES_DATETIME_FROM_ISOFORMAT.extend(
            [
                (
                    '2017-09-05 12:13:10+02:00',  # DST
                    datetime(2017, 9, 5, 12, 13, 10, 0,
                             ZoneInfo('Europe/Berlin'))
                ),
            ]
        )


@pytest.mark.parametrize(
    "dt_str, exp_dt",
    TESTCASES_DATETIME_FROM_ISOFORMAT)
def test_datetime_from_isoformat(dt_str, exp_dt):
    """
    Test function for datetime_from_isoformat().
    """

    # The function to be tested
    dt = datetime_from_isoformat(dt_str)

    # We need to compare the UTC times, because dateutil.parse() uses a
    # different tzinfo implementation than ZoneInfo.
    dt_utc = dt.astimezone(timezone.utc)
    exp_dt_utc = exp_dt.astimezone(timezone.utc)
    assert dt_utc == exp_dt_utc


def test_tzlocal():
    """
    Test function for tzlocal().
    """

    # The function to be tested
    tz = tzlocal()

    assert isinstance(tz, tzinfo)
    assert hasattr(tz, 'utcoffset')


def cpc_for_filtering():
    """
    Return a mocked zhmcclient.Cpc object for testing filter matching.

    Returns:
      zhmcclient.Cpc: Mocked CPC.
    """
    session = FakedSession('fake-host', 'fake-hmc', '2.13.1', '1.8')
    client = Client(session)
    session.hmc.cpcs.add({
        'object-id': 'fake-cpc1-oid',
        # object-uri is set up automatically
        'parent': None,
        'class': 'cpc',
        'name': 'fake-cpc1-name',
        'description': 'CPC #1 (DPM mode)',
        'status': 'active',
        'dpm-enabled': True,

        # Artificial properties for filter matching. The zhmcclient mock support
        # does not care about the actual properties, so we can add artificial
        # properties for the purpose of testing.
        'filter_none': None,
        'filter_bool_true': True,
        'filter_bool_false': False,
        'filter_int_0': 0,
        'filter_int_42': 42,
        'filter_float_0_0': 0.0,
        'filter_float_42_0': 42.0,
        'filter_str_empty': '',
        'filter_str_abc': 'abc',
    })
    cpc = client.cpcs.find(name='fake-cpc1-name')
    return cpc


CPC_FOR_FILTERING = cpc_for_filtering()

TESTCASES_MATCHES_FILTERS = [
    # Test cases for test_matches_filters().
    # Each list item is a tuple defining a testcase in the following format:
    # - desc (str): Testcase description
    # - obj (zhmcclient.BaseResource): Input object
    # - filter_args (dict): Input filter args
    # - exp_result (bool): Expected method return value
    # - exp_exc_type: Expected exception type raised from method to be tested,
    #   or None for success.
    # - exp_exc_pattern: Regex pattern to check exception message,
    #   or None for success.
    (
        "Filter args with non-existing property",
        CPC_FOR_FILTERING,
        {'filter_non_existing': 'x'},
        False,
        None,
        None,
    ),
    (
        "Filter args None",
        CPC_FOR_FILTERING,
        None,
        True,
        None,
        None,
    ),
    (
        "Filter args empty dict",
        CPC_FOR_FILTERING,
        {},
        True,
        None,
        None,
    ),
    (
        "filter_none with matching value",
        CPC_FOR_FILTERING,
        {'filter_none': None},
        True,
        None,
        None,
    ),
    (
        "filter_none with non-matching int value",
        CPC_FOR_FILTERING,
        {'filter_none': 0},
        False,
        None,
        None,
    ),
    (
        "filter_none with non-matching empty str value",
        CPC_FOR_FILTERING,
        {'filter_none': ''},
        False,
        None,
        None,
    ),
    (
        "filter_bool_true with non-convertible str",
        CPC_FOR_FILTERING,
        {'filter_bool_true': 'x'},
        None,
        FilterConversionError,
        "Cannot convert match value .* to bool",
    ),
    (
        "filter_bool_true with matching value of same type",
        CPC_FOR_FILTERING,
        {'filter_bool_true': True},
        True,
        None,
        None,
    ),
    (
        "filter_bool_true with matching value of int type",
        CPC_FOR_FILTERING,
        {'filter_bool_true': 1},
        True,
        None,
        None,
    ),
    (
        "filter_bool_true with matching value of str type, mixed case",
        CPC_FOR_FILTERING,
        {'filter_bool_true': 'tRuE'},
        True,
        None,
        None,
    ),
    (
        "filter_bool_true with non-matching value of same type",
        CPC_FOR_FILTERING,
        {'filter_bool_true': False},
        False,
        None,
        None,
    ),
    (
        "filter_bool_true with non-matching value of int type",
        CPC_FOR_FILTERING,
        {'filter_bool_true': 0},
        False,
        None,
        None,
    ),
    (
        "filter_bool_true with non-matching value of str type, mixed case",
        CPC_FOR_FILTERING,
        {'filter_bool_true': 'fAlSe'},
        False,
        None,
        None,
    ),
    (
        "filter_bool_false with matching value of same type",
        CPC_FOR_FILTERING,
        {'filter_bool_false': False},
        True,
        None,
        None,
    ),
    (
        "filter_bool_false with matching value of int type",
        CPC_FOR_FILTERING,
        {'filter_bool_false': 0},
        True,
        None,
        None,
    ),
    (
        "filter_bool_false with matching value of str type, mixed case",
        CPC_FOR_FILTERING,
        {'filter_bool_false': 'fAlSe'},
        True,
        None,
        None,
    ),
    (
        "filter_bool_false with non-matching value of same type",
        CPC_FOR_FILTERING,
        {'filter_bool_false': True},
        False,
        None,
        None,
    ),
    (
        "filter_bool_false with non-matching value of int type",
        CPC_FOR_FILTERING,
        {'filter_bool_false': 1},
        False,
        None,
        None,
    ),
    (
        "filter_bool_false with non-matching value of str type, mixed case",
        CPC_FOR_FILTERING,
        {'filter_bool_false': 'tRuE'},
        False,
        None,
        None,
    ),
    (
        "filter_int_0 with non-convertible str",
        CPC_FOR_FILTERING,
        {'filter_int_0': 'x'},
        None,
        FilterConversionError,
        "Cannot convert match value .* to int",
    ),
    (
        "filter_int_0 with matching value of same type",
        CPC_FOR_FILTERING,
        {'filter_int_0': 0},
        True,
        None,
        None,
    ),
    (
        "filter_int_0 with matching value of bool type",
        CPC_FOR_FILTERING,
        {'filter_int_0': False},
        True,
        None,
        None,
    ),
    (
        "filter_int_0 with matching value of str type",
        CPC_FOR_FILTERING,
        {'filter_int_0': '0'},
        True,
        None,
        None,
    ),
    (
        "filter_int_0 with non-matching value of same type",
        CPC_FOR_FILTERING,
        {'filter_int_0': 1},
        False,
        None,
        None,
    ),
    (
        "filter_int_0 with non-matching value of bool type",
        CPC_FOR_FILTERING,
        {'filter_int_0': True},
        False,
        None,
        None,
    ),
    (
        "filter_int_0 with non-matching value of str type",
        CPC_FOR_FILTERING,
        {'filter_int_0': '1'},
        False,
        None,
        None,
    ),
    (
        "filter_int_42 with matching value of same type",
        CPC_FOR_FILTERING,
        {'filter_int_42': 42},
        True,
        None,
        None,
    ),
    (
        "filter_int_42 with matching value of str type",
        CPC_FOR_FILTERING,
        {'filter_int_42': '42'},
        True,
        None,
        None,
    ),
    (
        "filter_int_42 with non-matching value of same type",
        CPC_FOR_FILTERING,
        {'filter_int_42': 7},
        False,
        None,
        None,
    ),
    (
        "filter_int_42 with non-matching value of str type",
        CPC_FOR_FILTERING,
        {'filter_int_42': '7'},
        False,
        None,
        None,
    ),
    (
        "filter_float_0_0 with non-convertible str",
        CPC_FOR_FILTERING,
        {'filter_float_0_0': 'x'},
        None,
        FilterConversionError,
        "Cannot convert match value .* to float",
    ),
    (
        "filter_float_0_0 with matching value of same type",
        CPC_FOR_FILTERING,
        {'filter_float_0_0': 0.0},
        True,
        None,
        None,
    ),
    (
        "filter_float_0_0 with matching value of str type",
        CPC_FOR_FILTERING,
        {'filter_float_0_0': '0.0'},
        True,
        None,
        None,
    ),
    (
        "filter_float_0_0 with non-matching value of same type",
        CPC_FOR_FILTERING,
        {'filter_float_0_0': 7.0},
        False,
        None,
        None,
    ),
    (
        "filter_float_0_0 with non-matching value of str type",
        CPC_FOR_FILTERING,
        {'filter_float_0_0': '7.0'},
        False,
        None,
        None,
    ),
    (
        "filter_float_42_0 with matching value of same type",
        CPC_FOR_FILTERING,
        {'filter_float_42_0': 42.0},
        True,
        None,
        None,
    ),
    (
        "filter_float_42_0 with matching value of int type",
        CPC_FOR_FILTERING,
        {'filter_float_42_0': 42},
        True,
        None,
        None,
    ),
    (
        "filter_float_42_0 with non-matching value of same type",
        CPC_FOR_FILTERING,
        {'filter_float_42_0': 7.0},
        False,
        None,
        None,
    ),
    (
        "filter_float_42_0 with non-matching value of str type",
        CPC_FOR_FILTERING,
        {'filter_float_42_0': '7.0'},
        False,
        None,
        None,
    ),
    (
        "filter_str_empty with matching value of same type",
        CPC_FOR_FILTERING,
        {'filter_str_empty': ''},
        True,
        None,
        None,
    ),
    (
        "filter_str_empty with matching regexp pattern",
        CPC_FOR_FILTERING,
        {'filter_str_empty': 'a?'},
        True,
        None,
        None,
    ),
    (
        "filter_str_empty with matching list of values",
        CPC_FOR_FILTERING,
        {'filter_str_empty': ['a', '']},
        True,
        None,
        None,
    ),
    (
        "filter_str_empty with matching list of regexp patterns",
        CPC_FOR_FILTERING,
        {'filter_str_empty': ['a', 'a?']},
        True,
        None,
        None,
    ),
    (
        "filter_str_empty with non-matching value of same type",
        CPC_FOR_FILTERING,
        {'filter_str_empty': 'a'},
        False,
        None,
        None,
    ),
    (
        "filter_str_empty with non-matching regexp pattern",
        CPC_FOR_FILTERING,
        {'filter_str_empty': 'ab?'},
        False,
        None,
        None,
    ),
    (
        "filter_str_empty with non-matching list of values",
        CPC_FOR_FILTERING,
        {'filter_str_empty': ['a', 'b']},
        False,
        None,
        None,
    ),
    (
        "filter_str_abc with matching value of same type",
        CPC_FOR_FILTERING,
        {'filter_str_abc': 'abc'},
        True,
        None,
        None,
    ),
    (
        "filter_str_abc with matching regexp pattern",
        CPC_FOR_FILTERING,
        {'filter_str_abc': 'ab?[c]'},
        True,
        None,
        None,
    ),
    (
        "filter_str_abc with matching list of values",
        CPC_FOR_FILTERING,
        {'filter_str_abc': ['a', 'abc']},
        True,
        None,
        None,
    ),
    (
        "filter_str_abc with matching list of regexp patterns",
        CPC_FOR_FILTERING,
        {'filter_str_abc': ['a?', 'ab?[c]']},
        True,
        None,
        None,
    ),
    (
        "filter_str_abc with non-matching value of same type",
        CPC_FOR_FILTERING,
        {'filter_str_abc': 'a'},
        False,
        None,
        None,
    ),
    (
        "filter_str_abc with non-matching regexp pattern",
        CPC_FOR_FILTERING,
        {'filter_str_abc': 'ab?'},
        False,
        None,
        None,
    ),
    (
        "filter_str_abc with non-matching list of values",
        CPC_FOR_FILTERING,
        {'filter_str_abc': ['a', 'b']},
        False,
        None,
        None,
    ),
    (
        "filter_str_abc with non-matching list of regexp patterns",
        CPC_FOR_FILTERING,
        {'filter_str_abc': ['a?', 'ab?']},
        False,
        None,
        None,
    ),
    (
        "Two filter args, both matching",
        CPC_FOR_FILTERING,
        {
            'filter_str_abc': 'abc',
            'filter_int_42': 42,
        },
        True,
        None,
        None,
    ),
    (
        "Two filter args, only one first one matching",
        CPC_FOR_FILTERING,
        {
            'filter_str_abc': 'abc',
            'filter_int_42': 0,
        },
        False,
        None,
        None,
    ),
    (
        "Two filter args, only one second one matching",
        CPC_FOR_FILTERING,
        {
            'filter_str_abc': 'x',
            'filter_int_42': 42,
        },
        False,
        None,
        None,
    ),
]


@pytest.mark.parametrize(
    "desc, obj, filter_args, exp_result, exp_exc_type, exp_exc_pattern",
    TESTCASES_MATCHES_FILTERS)
def test_matches_filters(
        desc, obj, filter_args, exp_result, exp_exc_type, exp_exc_pattern):
    # pylint: disable=unused-argument
    """
    Test function for matches_filters().
    """

    if exp_exc_type:
        with pytest.raises(exp_exc_type) as exc_info:

            # Execute the code to be tested
            matches_filters(obj, filter_args)

        exc = exc_info.value
        assert re.search(exp_exc_pattern, str(exc))
    else:

        # The function to be tested
        result = matches_filters(obj, filter_args)

        assert result == exp_result


TESTCASES_DIVIDE_FILTER_ARGS = [
    # Test cases for test_divide_filter_args().
    # Each list item is a tuple defining a testcase in the following format:
    # - desc (str): Testcase description
    # - query_props (list of str): Names of properties for server-side filtering
    # - filter_args (dict): Input filter args
    # - exp_result (tuple (query_parms, client_filter_args): Expected method
    #   return value
    # - exp_exc_type: Expected exception type raised from method to be tested,
    #   or None for success.
    # - exp_exc_pattern: Regex pattern to check exception message,
    #   or None for success.
    (
        "Filter args None, with empty query props",
        [],
        None,
        (
            [],
            {},
        ),
        None,
        None,
    ),
    (
        "Filter args None, with query props",
        ['prop1'],
        None,
        (
            [],
            {},
        ),
        None,
        None,
    ),
    (
        "Filter args for client-side with value None",
        ['prop2'],
        {'prop1': None},
        (
            [],
            {'prop1': None},
        ),
        None,
        None,
    ),
    (
        "Filter args for client-side with bool True",
        ['prop2'],
        {'prop1': True},
        (
            [],
            {'prop1': True},
        ),
        None,
        None,
    ),
    (
        "Filter args for client-side with bool False",
        ['prop2'],
        {'prop1': False},
        (
            [],
            {'prop1': False},
        ),
        None,
        None,
    ),
    (
        "Filter args for client-side with int",
        ['prop2'],
        {'prop1': 42},
        (
            [],
            {'prop1': 42},
        ),
        None,
        None,
    ),
    (
        "Filter args for client-side with float",
        ['prop2'],
        {'prop1': 42.0},
        (
            [],
            {'prop1': 42.0},
        ),
        None,
        None,
    ),
    (
        "Filter args for client-side with empty str",
        ['prop2'],
        {'prop1': ''},
        (
            [],
            {'prop1': ''},
        ),
        None,
        None,
    ),
    (
        "Filter args for client-side with non-empty str",
        ['prop2'],
        {'prop1': 'x'},
        (
            [],
            {'prop1': 'x'},
        ),
        None,
        None,
    ),
    (
        "Filter args for server-side with value None",
        ['prop1'],
        {'prop1': None},
        (
            ['prop1=None'],
            {},
        ),
        None,
        None,
    ),
    (
        "Filter args for server-side with bool True",
        ['prop1'],
        {'prop1': True},
        (
            ['prop1=True'],
            {},
        ),
        None,
        None,
    ),
    (
        "Filter args for server-side with bool False",
        ['prop1'],
        {'prop1': False},
        (
            ['prop1=False'],
            {},
        ),
        None,
        None,
    ),
    (
        "Filter args for server-side with int",
        ['prop1'],
        {'prop1': 42},
        (
            ['prop1=42'],
            {},
        ),
        None,
        None,
    ),
    (
        "Filter args for server-side with float",
        ['prop1'],
        {'prop1': 42.0},
        (
            ['prop1=42.0'],
            {},
        ),
        None,
        None,
    ),
    (
        "Filter args for server-side with empty str",
        ['prop1'],
        {'prop1': ''},
        (
            ['prop1='],
            {},
        ),
        None,
        None,
    ),
    (
        "Filter args for server-side with non-empty str",
        ['prop1'],
        {'prop1': 'x'},
        (
            ['prop1=x'],
            {},
        ),
        None,
        None,
    ),
]


@pytest.mark.parametrize(
    "desc, query_props, filter_args, exp_result, exp_exc_type, exp_exc_pattern",
    TESTCASES_DIVIDE_FILTER_ARGS)
def test_divide_filter_args(
        desc, query_props, filter_args, exp_result, exp_exc_type,
        exp_exc_pattern):
    # pylint: disable=unused-argument
    """
    Test function for divide_filter_args().
    """

    if exp_exc_type:
        with pytest.raises(exp_exc_type) as exc_info:

            # Execute the code to be tested
            divide_filter_args(query_props, filter_args)

        exc = exc_info.value
        assert re.search(exp_exc_pattern, str(exc))
    else:

        # The function to be tested
        result = divide_filter_args(query_props, filter_args)

        assert result == exp_result



# Copyright 2019,2021 IBM Corp. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""
Pytest fixtures for CPCs.
"""


import pytest
import zhmcclient

# pylint: disable=unused-import
from ._hmc_definition_fixtures import hmc_session  # noqa: F401
from ._hmc_inventory_file import HMCInventoryFileError


__all__ = ['all_cpcs', 'dpm_mode_cpcs', 'classic_mode_cpcs']


def fixtureid_cpcs(fixture_value):
    """
    Return a fixture ID to be used by pytest, for fixtures returning a list of
    CPCs.

    Parameters:

      * fixture_value (list of zhmcclient.Cpc): The list of CPCs.
    """
    if not isinstance(fixture_value, list):
        return None  # Use pytest auto-generated ID
    cpc_list = fixture_value
    cpcs_str = ','.join([cpc.name for cpc in cpc_list])
    return f"CPCs={cpcs_str}"


@pytest.fixture(
    scope='module',
    ids=fixtureid_cpcs
)
def all_cpcs(request, hmc_session):  # noqa: F811
    # pylint: disable=redefined-outer-name,unused-argument
    """
    Pytest fixture representing the set of all CPCs to test against, regardless
    of their operational mode.

    The set of CPCs to test against is defined in the ``cpcs`` variable in the
    HMC entry in the :ref:`HMC inventory file`; if that variable is omitted, it
    is the set of all managed CPCs.

    A test function parameter with the name of this fixture resolves to a list
    of :class:`zhmcclient.Cpc` objects representing that set of CPCs. These
    objects have the "short" set of properties from :meth:`zhmcclient.Cpc.list`.

    The test function is invoked just once with the list of CPCs, and the
    test function needs to loop through the CPCs (or a subset).

    Because the `hmc_session` parameter of this fixture is again a fixture,
    the :func:`zhmcclient.testutils.hmc_session`
    function needs to be imported as well when this fixture is used.
    """
    return defined_cpcs(hmc_session, 'any')


@pytest.fixture(
    scope='module',
    ids=fixtureid_cpcs
)
def dpm_mode_cpcs(request, hmc_session):  # noqa: F811
    # pylint: disable=redefined-outer-name,unused-argument
    """
    Pytest fixture representing the set of CPCs in DPM mode to test against.

    The set of CPCs to test against is the subset of all CPCs to be tested that
    is in DPM mode.

    The set of all CPCs to be tested is defined in the ``cpcs`` variable in the
    HMC entry in the :ref:`HMC inventory file`; if that variable is omitted, it
    is the set of all managed CPCs.

    A test function parameter with the name of this fixture resolves to a list
    of :class:`zhmcclient.Cpc` objects representing that set of CPCs. These
    objects have the "short" set of properties from :meth:`zhmcclient.Cpc.list`.

    The test function is invoked just once with the list of CPCs, and the
    test function needs to loop through the CPCs (or a subset).

    Because the `hmc_session` parameter of this fixture is again a fixture,
    the :func:`zhmcclient.testutils.hmc_session`
    function needs to be imported as well when this fixture is used.
    """
    return defined_cpcs(hmc_session, 'dpm')


@pytest.fixture(
    scope='module',
    ids=fixtureid_cpcs
)
def classic_mode_cpcs(request, hmc_session):  # noqa: F811
    # pylint: disable=redefined-outer-name,unused-argument
    """
    Pytest fixture representing the set of CPCs in classic mode to test against.

    The set of CPCs to test against is the subset of all CPCs to be tested that
    is in classic mode.

    The set of all CPCs to be tested is defined in the ``cpcs`` variable in the
    HMC entry in the :ref:`HMC inventory file`; if that variable is omitted, it
    is the set of all managed CPCs.

    A test function parameter with the name of this fixture resolves to a list
    of :class:`zhmcclient.Cpc` objects representing that set of CPCs. These
    objects have the "short" set of properties from :meth:`zhmcclient.Cpc.list`.

    The test function is invoked just once with the list of CPCs, and the
    test function needs to loop through the CPCs (or a subset).

    Because the `hmc_session` parameter of this fixture is again a fixture,
    the :func:`zhmcclient.testutils.hmc_session`
    function needs to be imported as well when this fixture is used.
    """
    return defined_cpcs(hmc_session, 'classic')


def defined_cpcs(session, mode):
    """
    Return a list of CPCs defined in the HMC definition file, that are
    managed by the HMC, and that have the desired operational mode.

    Parameters:

      session (:class:`zhmcclient.Session`): The session with the HMC.

      mode (string): The desired mode of the CPC ('dpm', 'classic', 'any').

    Returns:

      list of :class:`zhmcclient.Cpc`: The CPCs in the desired mode.
    """
    client = zhmcclient.Client(session)
    managed_cpcs = client.cpcs.list()
    managed_cpc_names = [cpc.name for cpc in managed_cpcs]
    hd = session.hmc_definition
    test_cpc_names = list(hd.cpcs.keys()) if hd.cpcs else managed_cpc_names

    result_cpcs = []
    for cpc_name in test_cpc_names:

        cpcs = [cpc for cpc in managed_cpcs if cpc.name == cpc_name]
        if not cpcs:
            raise HMCInventoryFileError(
                f"CPC {cpc_name} defined for HMC {hd.nickname} in the HMC "
                "inventory file is not managed by that HMC. "
                f"Managed CPCs: {', '.join(managed_cpc_names)}")
        cpc = cpcs[0]

        if hd.cpcs and cpc_name in hd.cpcs:
            hd_props = hd.cpcs[cpc_name]
            hd_dpm_enabled = hd_props.get('dpm_enabled', None)
            if hd_dpm_enabled is not None and hd_dpm_enabled != cpc.dpm_enabled:
                raise HMCInventoryFileError(
                    f"CPC {cpc_name} defined for HMC {hd.nickname} in the HMC "
                    f"inventory file defines dpm_enabled={hd_dpm_enabled}, "
                    f"but the actual CPC has dpm_enabled={cpc.dpm_enabled}.")

        if mode == 'dpm' and not cpc.dpm_enabled:
            continue
        if mode == 'classic' and cpc.dpm_enabled:
            continue
        result_cpcs.append(cpc)

    return result_cpcs



# Copyright 2024 IBM Corp. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""
A :term:`Partition Link` interconnects two or more
:term:`Partitions <Partition>` that share the same network configuration.

Partition Links have been introduced with SE 2.16. Conceptually, Partition Links
can connect different :term:`CPCs <CPC>`. However, the implementation is
currently limited to the same DPM-enabled CPC. Currently, one Partition Link
resource is always associated with a single CPC at any time.

Partition Links are based on different technologies, indicated in their
``type`` property:

* ``smc-d`` - Shared Memory Communications - Direct Memory Access (SMC-D)
  (Version 2 or later).
  Support for this technology is indicated via :ref:`API feature <API features>`
  "dpm-smcd-partition-link-management".
* ``hipersockets`` - Hipersockets.
  Support for this technology is indicated via :ref:`API feature <API features>`
  "dpm-hipersockets-partition-link-management".
* ``ctc`` - FICON Channel to Channel (CTC) interconnect, using corresponding
  cabling between FICON adapters.
  Support for this technology is indicated via :ref:`API feature <API features>`
  "dpm-ctc-partition-link-management".

Partition Links have a ``state`` property that indicates their current
attachment state to any partitions:

* ``complete`` - All requests for attaching or detaching the partition link to
  or from partitions are complete, and the partition link is attached to 2 or
  more partitions.
* ``incomplete`` - All requests for attaching or detaching the partition link to
  or from partitions are complete, but the partition link is attached to less
  than 2 partitions.
* ``updating`` - Some requests for attaching or detaching the partition link to
  or from partitions are incomplete.

This section describes the interface for Partition Links using resource class
:class:`~zhmcclient.PartitionLink` and the corresponding manager class
:class:`~zhmcclient.PartitionLinkManager`.

Because conceptually, Partition Links can connect different CPCs, this client
has designed :class:`~zhmcclient.PartitionLink` resources to be available
through the :class:`~zhmcclient.Console` resource, via its
:attr:`~zhmcclient.Console.partition_links` property.

The earlier interfaces for Hipersockets are also supported:

* represented as :class:`~zhmcclient.Adapter`, including support for creation
  and deletion.
* Attachment to a partition is managed via creation and deletion of
  :class:`~zhmcclient.Nic` resources on the partition.
"""

import copy
import re
import time

from ._manager import BaseManager
from ._resource import BaseResource
from ._logging import get_logger, logged_api_call
from ._constants import HMC_LOGGER_NAME
from ._utils import RC_PARTITION_LINK
from ._exceptions import PartitionLinkError, OperationTimeout

__all__ = ['PartitionLinkManager', 'PartitionLink']

HMC_LOGGER = get_logger(HMC_LOGGER_NAME)


class PartitionLinkManager(BaseManager):
    """
    Manager providing access to the :term:`partition links <partition link>` of
    the HMC.

    Derived from :class:`~zhmcclient.BaseManager`; see there for common methods
    and attributes.

    Objects of this class are not directly created by the user; they are
    accessible via the following instance variable:

    * :attr:`~zhmcclient.Console.partition_links` of a
      :class:`~zhmcclient.Console` object.

    HMC/SE version requirements:

    * SE version >= 2.16.0
    * for technology-specific support, see the API features described in
      :ref:`Partition Links`
    """

    def __init__(self, console):
        # This function should not go into the docs.
        # Parameters:
        #   console (:class:`~zhmcclient.Console`):
        #     CPC or HMC defining the scope for this manager.

        # Resource properties that are supported as filter query parameters.
        # If the support for a resource property changes within the set of HMC
        # versions that support this type of resource, this list must be set up
        # for the version of the HMC this session is connected to.
        query_props = [
            'cpc-uri',
            'name',
            'state',
        ]

        super().__init__(
            resource_class=PartitionLink,
            class_name=RC_PARTITION_LINK,
            session=console.manager.session,
            parent=console,
            base_uri='/api/partition-links',
            oid_prop='object-id',
            uri_prop='object-uri',
            name_prop='name',
            query_props=query_props)
        self._console = console

    @property
    def console(self):
        """
        :class:`~zhmcclient.Console`: The Console object representing the HMC.
        """
        return self._console

    @logged_api_call
    def list(self, full_properties=False, filter_args=None,
             additional_properties=None):
        """
        List the partition links known to the HMC.

        Partition links for which the authenticated user does not have
        object-access permission are not included.

        Any resource property may be specified in a filter argument. For
        details about filter arguments, see :ref:`Filtering`.

        The listing of resources is handled in an optimized way:

        * If this manager is enabled for :ref:`auto-updating`, a locally
          maintained resource list is used (which is automatically updated via
          inventory notifications from the HMC) and the provided filter
          arguments are applied.

        * Otherwise, if the filter arguments specify the resource name as a
          single filter argument with a straight match string (i.e. without
          regular expressions), an optimized lookup is performed based on a
          locally maintained name-URI cache.

        * Otherwise, the HMC List operation is performed with the subset of the
          provided filter arguments that can be handled on the HMC side and the
          remaining filter arguments are applied on the client side on the list
          result.

        This method performs the "List Partition Links" HMC operation.

        HMC/SE version requirements:

        * SE version >= 2.16.0
        * for technology-specific support, see the API features described in
          :ref:`Partition Links`

        Authorization requirements:

        * Object-access permission to any partition links to be included in the
          result.

        Parameters:

          full_properties (bool):
            Controls that the full set of resource properties for each returned
            partition link is being retrieved, vs. only the following short
            set: "object-uri", "cpc-uri", "name", "state", and "type".

          filter_args (dict):
            Filter arguments that narrow the list of returned resources to
            those that match the specified filter arguments. For details, see
            :ref:`Filtering`.

            `None` causes no filtering to happen.

          additional_properties (list of string):
            List of property names that are to be returned in addition to the
            short set of properties.

        Returns:

          : A list of :class:`~zhmcclient.PartitionLink` objects.

        Raises:

          :exc:`~zhmcclient.HTTPError`
          :exc:`~zhmcclient.ParseError`
          :exc:`~zhmcclient.AuthError`
          :exc:`~zhmcclient.ConnectionError`
          :exc:`~zhmcclient.FilterConversionError`
        """
        result_prop = 'partition-links'
        list_uri = self._base_uri
        return self._list_with_operation(
            list_uri, result_prop, full_properties, filter_args,
            additional_properties)

    @logged_api_call
    def create(self, properties=None):
        """
        Create a partition link from the input properties.

        The input properties may specify an initial attachment of the partition
        link to one or more partitions.

        Additional attachments to partitions and detachments from partitions
        can be performed with the following :class:`~zhmcclient.Partition`
        methods:

        * :meth:`zhmcclient.Partition.attach_network_link`
        * :meth:`zhmcclient.Partition.detach_network_link`
        * :meth:`zhmcclient.Partition.attach_ctc_link`
        * :meth:`zhmcclient.Partition.detach_ctc_link`

        and also with the :meth:`zhmcclient.PartitionLink.update_properties`
        method.

        For CTC-type partition links, it is required to specify the paths
        describing the physical connectivity between the FICON adapters at
        creation time, using the ``paths`` input property.

        The new partition link will be associated with the CPC identified by the
        ``cpc-uri`` input property.

        This method performs the "Create Partition Link" HMC operation and
        waits for completion of its asynchronous job.

        HMC/SE version requirements:

        * SE version >= 2.16.0
        * for technology-specific support, see the API features described in
          :ref:`Partition Links`

        Authorization requirements:

        * Object-access permission to the CPC that will be associated with
          the new partition link.
        * Task permission to the "Create Partition Link" task.
        * Object-access permission to all Partitions for the initially requested
          attachments.
        * Object-access permission to all FICON adapter objects used for the
          CTC connections for the initially requested attachments.

        Parameters:

          properties (dict): Initial property values.
            Allowable properties are defined in section 'Request body contents'
            in section 'Create Partition Link' in the :term:`HMC API` book.

        Returns:

          :class:`~zhmcclient.PartitionLink`: The new partition link.

        Raises:

          :exc:`~zhmcclient.HTTPError`
          :exc:`~zhmcclient.ParseError`
          :exc:`~zhmcclient.AuthError`
          :exc:`~zhmcclient.ConnectionError`
          :exc:`~zhmcclient.OperationTimeout`: The timeout expired while
            waiting for completion of the operation.

        Examples:

        * Create Hipersockets-type partition link with no attached partitions::

            part_link = console.partition_links.create({
                "name": "mylink",
                "type": "hipersockets",
                "cpc-uri": cpc.uri
            })

        * Create CTC-type partition link with one path and two attached
          partitions::

            part_link = console.partition_links.create({
                "name": "mylink",
                "type": "ctc",
                "cpc-uri": cpc.uri,
                "partitions": [part1.uri, part2.uri],
                "paths": [
                    {
                        "adapter-port-uri": fc_port1.uri,
                        "connecting-adapter-port-uri": fc_port2.uri
                    }
                ]
            })

        """
        if properties is None:
            properties = {}

        result = self.session.post(
            uri=self._base_uri,
            body=properties,
            wait_for_completion=True)

        # The "Create Partition Link" operation does not return the object-uri
        # of the new partition link in the response. however, it does return
        # it in the "Location" header, and the session.post() method adds that
        # to its return value as the "location-uri" field.
        uri = result['location-uri']
        name = properties[self._name_prop]
        props = copy.deepcopy(properties)
        # Remove input parameters that are not 1:1 in the data model
        if 'partitions' in props:
            del props['partitions']  # not in data model
        if 'paths' in props:
            del props['paths']  # different structure in data model
        props[self._uri_prop] = uri
        partition_link = PartitionLink(self, uri, name, props)
        self._name_uri_cache.update(name, uri)
        return partition_link


class PartitionLink(BaseResource):
    """
    Representation of a :term:`partition link`.

    Derived from :class:`~zhmcclient.BaseResource`; see there for common
    methods and attributes.

    Objects of this class are not directly created by the user; they are
    returned from creation or list functions on their manager object
    (in this case, :class:`~zhmcclient.PartitionLinkManager`).

    HMC/SE version requirements:

    * SE version >= 2.16.0
    * for technology-specific support, see the API features described in
      :ref:`Partition Links`
    """

    def __init__(self, manager, uri, name=None, properties=None):
        # This function should not go into the docs.
        #   manager (:class:`~zhmcclient.PartitionLinkManager`):
        #     Manager object for this resource object.
        #   uri (string):
        #     Canonical URI path of the resource.
        #   name (string):
        #     Name of the resource.
        #   properties (dict):
        #     Properties to be set for this resource object. May be `None` or
        #     empty.
        assert isinstance(manager, PartitionLinkManager), (
            "PartitionLink init: Expected manager type "
            f"{PartitionLinkManager}, got {type(manager)}")
        super().__init__(manager, uri, name, properties)
        # The manager objects for child resources (with lazy initialization):
        self._cpc = None

    @property
    def cpc(self):
        """
        :class:`~zhmcclient.Cpc`: The :term:`CPC` to which this partition link
        is associated.

        The returned :class:`~zhmcclient.Cpc` has only a minimal set of
        properties populated.
        """
        # We do here some lazy loading.
        if not self._cpc:
            cpc_uri = self.get_property('cpc-uri')
            cpc_mgr = self.manager.console.manager.client.cpcs
            self._cpc = cpc_mgr.resource_object(cpc_uri)
        return self._cpc

    @logged_api_call
    def list_attached_partitions(self, name=None, status=None):
        """
        Return the partitions to which this partition link is currently
        attached, optionally filtered by partition name and status.

        Properties of the endpoint objects of the partition links are not
        returned by this method and can be obtained via the
        :class:`~zhmcclient.PartitionLink` properties.
        In case of Hipersockets, these properties are also available
        on the NIC objects. In case of SMC-D and CTC, the HMC API does not
        externalize corresponding endpoint objects other than as properties
        of partition link objexts.

        This method performs the "Get Partition Link Properties" HMC operation,
        if this Python object does not yet have the full properties.

        HMC/SE version requirements:

        * SE version >= 2.16.0
        * for technology-specific support, see the API features described in
          :ref:`Partition Links`

        Authorization requirements:

        * Object-access permission to this partition link.

        Parameters:

          name (:term:`string`): Filter pattern (regular expression)
            to limit returned partitions to those that have a matching
            name. If `None`, no filtering for the partition name takes place.

          status (:term:`string`): Filter string to limit returned
            partitions to those that have a matching status. The value
            must be a valid partition 'status' property value. If `None`, no
            filtering for the partition status takes place.

        Returns:

          List of :class:`~zhmcclient.Partition` objects representing the
          partitions to which this partition link is currently attached, with
          a minimal set of properties ('object-id', 'name').

        Raises:

          :exc:`~zhmcclient.HTTPError`
          :exc:`~zhmcclient.ParseError`
          :exc:`~zhmcclient.AuthError`
          :exc:`~zhmcclient.ConnectionError`
        """
        part_list = []
        part_uri_list = []
        pl_type = self.get_property('type')
        if pl_type == 'ctc':
            for path in self.get_property('paths'):
                for device in path.get('devices'):
                    for endpoint in device.get('endpoint-pair'):
                        part_uri = endpoint.get('partition-uri')
                        part_name = endpoint.get('partition-name')
                        part = self.cpc.partitions.resource_object(
                            part_uri, {'name': part_name})
                        if name:
                            if not re.match(name, part_name):
                                continue
                        if status:
                            if part.get_property('status') != status:
                                continue
                        if part_uri not in part_uri_list:
                            part_uri_list.append(part_uri)
                            part_list.append(part)
        else:
            # Hipersockets or SMC-D
            for bc in self.get_property('bus-connections'):
                part_uri = bc.get('partition-uri')
                part_name = bc.get('partition-name')
                part = self.cpc.partitions.resource_object(
                    part_uri, {'name': part_name, 'nics': []})
                if name:
                    if not re.match(name, part_name):
                        continue
                if status:
                    if part.get_property('status') != status:
                        continue
                if part_uri not in part_uri_list:
                    part_uri_list.append(part_uri)
                    part_list.append(part)
        return part_list

    @logged_api_call
    def wait_for_states(self, states=('complete', 'incomplete'), timeout=30):
        """
        Wait for this partition link to reach one of the specified states.

        If there are changes in partition connectivity underway, the partition
        link' ``state`` property indicates whether the change is still going
        on or has reached a stable state, as follows:

          * "complete" - All partitions of this partition link are connected.
          * "incomplete" - Less than 2 partitions are connected to this
            partition link or an adapter is in degraded state.
          * "updating" - Partitions are currently being added or removed from
            this partition link.

        This method pulls the ``state`` property every 2 seconds until one of
        the specified states is reached. It times out after the specified
        timeout.

        HMC/SE version requirements:

        * SE version >= 2.16.0

        Authorization requirements:

        * Object-access permission to this partition link.

        Parameters:

          states (list of string): The states to wait for.

          timeout (int): The timeout in seconds. 0 means not to time out.

        Raises:

          :exc:`~zhmcclient.HTTPError`
          :exc:`~zhmcclient.ParseError`
          :exc:`~zhmcclient.AuthError`
          :exc:`~zhmcclient.ConnectionError`
          :exc:`~zhmcclient.OperationTimeout`: The timeout expired while
            waiting for completion of the operation.
        """

        if timeout > 0:
            start_time = time.time()

        while True:
            try:
                self.pull_properties(['state'])
                state = self.properties['state']
                if state in states:
                    return
            except ConnectionError:
                HMC_LOGGER.debug(
                    "Retrying after ConnectionError while waiting for states "
                    "%s in partition link %r (currently has state %s). This "
                    "could be because HMC is restarting.",
                    states, self.name, state)

            if timeout > 0:
                current_time = time.time()
                if current_time > start_time + timeout:
                    raise OperationTimeout(
                        f"Waiting for states {states} in partition link "
                        f"{self.name} timed out (timeout: {timeout} s, "
                        f"currently has state {state})",
                        timeout)

            time.sleep(2)  # Avoid hot spin loop

    @logged_api_call
    def delete(self, force_detach=False):
        """
        Delete this partition link on the HMC.

        If there are active partitions to which the partition link is attached,
        the operation will fail by default. The 'force_detach' parameter can be
        used to forcefully detach the partition link from active partitions
        before deleting it.

        This method performs the "Delete Partition Link" HMC operation and
        waits for completion of its asynchronous job.

        HMC/SE version requirements:

        * SE version >= 2.16.0
        * for technology-specific support, see the API features described in
          :ref:`Partition Links`

        Authorization requirements:

        * Object-access permission to this partition link.
        * Object-access permission to all partitions that currently have this
          partition link attached.
        * Task permission to the "Delete Partition Link" task.

        Parameters:

          force_detach (bool): Controls what to do with active partitions
            associated with this partition link. If True, such partitions
            are detached forcefully. If False, the operation fails with
            status code 409 (Conflict) and reason code 100.

        Raises:

          :exc:`~zhmcclient.HTTPError`
          :exc:`~zhmcclient.ParseError`
          :exc:`~zhmcclient.AuthError`
          :exc:`~zhmcclient.ConnectionError`
          :exc:`~zhmcclient.OperationTimeout`: The timeout expired while
            waiting for completion of the operation.
        """
        body = {}
        if force_detach:
            body['force-detach'] = True

        self.manager.session.post(
            uri=f'{self.uri}/operations/delete',
            resource=self,
            body=body,
            wait_for_completion=True,
            busy_retries=1, busy_wait=2)

        # pylint: disable=protected-access
        self.manager._name_uri_cache.delete(
            self.get_properties_local(self.manager._name_prop, None))
        self.cease_existence_local()

    @logged_api_call
    def update_properties(self, properties):
        """
        Update writeable properties of this partition link.

        This method can be used to attach and detach this partition link
        to and from partitions, by modifying the appropriate resource
        properties.

        This method serializes with other methods that access or change
        properties on the same Python object.

        This method performs the "Modify Partition Link" HMC operation and
        waits for completion of its asynchronous job.

        HMC/SE version requirements:

        * SE version >= 2.16.0
        * for technology-specific support, see the API features described in
          :ref:`Partition Links`

        Authorization requirements:

        * Task permission to the "Create Partition Link" task.
        * Object-access permission to all Partitions for the initially requested
          attachments.
        * Object-access permission to all FICON adapter objects used for the
          CTC connections for the initially requested attachments.

        Parameters:

          properties (dict): New values for the properties to be updated.
            Properties not to be updated are omitted.
            Allowable properties are listed for operation
            'Modify Partition Link' in section 'Partition Link object' in the
            :term:`HMC API` book.

        Raises:

          :exc:`~zhmcclient.HTTPError`
          :exc:`~zhmcclient.ParseError`
          :exc:`~zhmcclient.AuthError`
          :exc:`~zhmcclient.ConnectionError`
          :exc:`~zhmcclient.OperationTimeout`: The timeout expired while
            waiting for completion of the operation.
        """
        # pylint: disable=protected-access
        result = self.manager.session.post(
            uri=f'{self.uri}/operations/modify',
            resource=self,
            body=properties,
            wait_for_completion=True,
            busy_retries=1, busy_wait=2)

        for op_result in result['operation-results']:
            if op_result['operation-status'] \
                    not in ("attached", "detached"):
                raise PartitionLinkError(result['operation-results'])

        is_rename = self.manager._name_prop in properties
        if is_rename:
            # Delete the old name from the cache
            self.manager._name_uri_cache.delete(self.name)
        self.update_properties_local(copy.deepcopy(properties))
        if is_rename:
            # Add the new name to the cache
            self.manager._name_uri_cache.update(self.name, self.uri)

    def dump(self):
        """
        Dump this PartitionLink resource with its properties as a resource
        definition.

        The returned resource definition has the following format::

            {
                # Resource properties:
                "properties": {...},
            }

        Returns:

          dict: Resource definition of this resource.
        """

        # Dump the resource properties
        resource_dict = super().dump()

        return resource_dict



# Copyright 2024 IBM Corp. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""
End2end tests for partition links (on CPCs in DPM mode).

These tests do not change any existing partition links or partitions, but
create, modify and delete test partition links and test partitions.
"""


import random
import uuid
from copy import copy
from pprint import pprint
import pytest
from requests.packages import urllib3

import zhmcclient
# pylint: disable=line-too-long,unused-import
from zhmcclient.testutils import hmc_definition, hmc_session  # noqa: F401, E501
from zhmcclient.testutils import dpm_mode_cpcs  # noqa: F401, E501
# pylint: enable=line-too-long,unused-import

from .utils import skip_warn, pick_test_resources, TEST_PREFIX, \
    skipif_no_partition_link_feature, runtest_find_list, \
    runtest_get_properties, assert_properties, standard_partition_props

urllib3.disable_warnings()

# Properties in minimalistic PartitionLink objects (e.g. find_by_name())
PARTLINK_MINIMAL_PROPS = ['object-uri', 'name']

# Properties in PartitionLink objects returned by list() without full props
PARTLINK_LIST_PROPS = ['object-uri', 'cpc-uri', 'name', 'state', 'type']

# Properties whose values can change between retrievals of PartitionLink objs
PARTLINK_VOLATILE_PROPS = []


def replace_expressions(obj, replacements):
    """
    Return a deep copy of the input object, with any string values that are
    expressions containing replacement strings in 'replacements', with the
    corresponding values from 'replacements'.

    This is used to be able to specify adapters and partitions in the testcases
    as strings such as 'adapter_1' and then get them replaced with actual
    adapter or partition objects.

    The function is called recursively while walking through the initially
    provided properties dict.

    Parameters:
      obj (dict or object): The input object. Initially, a dict with properties.
      replacements (dict): Replacement values.

    Returns:
      object: A deep copy of the input object, with any replacements made.
    """
    if isinstance(obj, dict):
        ret_obj = {}
        for name, value in obj.items():
            ret_obj[name] = replace_expressions(value, replacements)
        return ret_obj

    if isinstance(obj, list):
        ret_obj = []
        for value in obj:
            ret_obj.append(replace_expressions(value, replacements))
        return ret_obj

    if isinstance(obj, str):
        ret_obj = obj
        if any(map(lambda rep: rep in obj, replacements.keys())):
            ret_obj = eval(f'{obj}', replacements)  # pylint: disable=eval-used
        return ret_obj

    return copy(obj)


@pytest.mark.parametrize(
    "pl_type", [
        'hipersockets',
        'smc-d',
        'ctc'
    ])
def test_partlink_find_list(dpm_mode_cpcs, pl_type):  # noqa: F811
    # pylint: disable=redefined-outer-name
    """
    Test list(), find(), findall().
    """
    if not dpm_mode_cpcs:
        pytest.skip("HMC definition does not include any CPCs in DPM mode")

    for cpc in dpm_mode_cpcs:
        assert cpc.dpm_enabled
        skipif_no_partition_link_feature(cpc)

        console = cpc.manager.client.consoles.console
        session = cpc.manager.session
        hd = session.hmc_definition

        # Pick the partition links to test with
        filter_args = {
            'cpc-uri': cpc.uri,
            'type': pl_type,
        }
        partlink_list = console.partition_links.list(filter_args=filter_args)
        if not partlink_list:
            skip_warn(f"No partition links of type {pl_type} associated to "
                      f"CPC {cpc.name} managed by HMC {hd.host}")
        partlink_list = pick_test_resources(partlink_list)

        for partlink in partlink_list:
            print(f"Testing on CPC {cpc.name} with {pl_type} partition link "
                  f"{partlink.name!r}")
            runtest_find_list(
                session, console.partition_links, partlink.name, 'name',
                'object-uri', PARTLINK_VOLATILE_PROPS, PARTLINK_MINIMAL_PROPS,
                PARTLINK_LIST_PROPS)


@pytest.mark.parametrize(
    "pl_type", [
        'hipersockets',
        'smc-d',
        'ctc'
    ])
def test_partlink_property(dpm_mode_cpcs, pl_type):  # noqa: F811
    # pylint: disable=redefined-outer-name
    """
    Test property related methods
    """
    if not dpm_mode_cpcs:
        pytest.skip("HMC definition does not include any CPCs in DPM mode")

    for cpc in dpm_mode_cpcs:
        assert cpc.dpm_enabled
        skipif_no_partition_link_feature(cpc)

        console = cpc.manager.client.consoles.console
        session = cpc.manager.session
        hd = session.hmc_definition

        # Pick the partition links to test with
        filter_args = {
            'cpc-uri': cpc.uri,
            'type': pl_type,
        }
        partlink_list = console.partition_links.list(filter_args=filter_args)
        if not partlink_list:
            skip_warn(f"No partition links of type {pl_type} associated to "
                      f"CPC {cpc.name} managed by HMC {hd.host}")
        partlink_list = pick_test_resources(partlink_list)

        for partlink in partlink_list:
            print(f"Testing on CPC {cpc.name} with {pl_type} partition link "
                  f"{partlink.name!r}")

            # Select a property that is not returned by list()
            non_list_prop = 'description'

            runtest_get_properties(partlink.manager, non_list_prop)


@pytest.mark.parametrize(
    "pl_type", [
        'hipersockets',
        'smc-d',
        'ctc'
    ])
def test_partlink_crud(dpm_mode_cpcs, pl_type):  # noqa: F811
    # pylint: disable=redefined-outer-name
    """
    Test create, read, update and delete a partition link.
    """
    if not dpm_mode_cpcs:
        pytest.skip("HMC definition does not include any CPCs in DPM mode")

    for cpc in dpm_mode_cpcs:
        assert cpc.dpm_enabled
        skipif_no_partition_link_feature(cpc)

        print(f"Testing on CPC {cpc.name}")

        console = cpc.manager.client.consoles.console
        partlink_name = f"{TEST_PREFIX}_crud_{uuid.uuid4().hex}"
        partlink_name_new = partlink_name + ' new'

        # Test creating the partition link.
        # For Hipersockets and SMC-D-type partition links, we create the
        # link with no partitions attached.
        # For CTC-type partition links, we create the link with one path
        # and two initial partitions.

        partlink_input_props = {
            'cpc-uri': cpc.uri,
            'name': partlink_name,
            'description': 'Test partition link for zhmcclient end2end tests',
            'type': pl_type,
        }

        if pl_type == 'ctc':

            adapters = cpc.adapters.list(
                filter_args={'type': 'fc', 'state': 'online'})
            if len(adapters) < 1:
                pytest.skip(f"CPC {cpc.name} has no online FC adapters "
                            "for CTC partition link creation")
            adapter = random.choice(adapters)

            part1_name = f"{TEST_PREFIX}_{uuid.uuid4().hex}"
            part1_props = standard_partition_props(cpc, part1_name)
            part1 = cpc.partitions.create(part1_props)
            part2_name = f"{TEST_PREFIX}_{uuid.uuid4().hex}"
            part2_props = standard_partition_props(cpc, part2_name)
            part2 = cpc.partitions.create(part2_props)

            path = {
                'adapter-port-uri': adapter.uri,
                'connecting-adapter-port-uri': adapter.uri,
            }
            partlink_input_props['partitions'] = [part1.uri, part2.uri]
            partlink_input_props['paths'] = [path]
            partlink_input_props['devices-per-path'] = 1

        partlink_auto_props = {
            'cpc-name': cpc.name,
        }

        partlink = None
        try:
            # The code to be tested
            partlink = console.partition_links.create(partlink_input_props)

            partlink.wait_for_states()

            # Remove input properties that are not in data model or that are
            # different in data model, so that we can check.
            if 'partitions' in partlink_input_props:
                del partlink_input_props['partitions']  # not in data model
            if 'paths' in partlink_input_props:
                del partlink_input_props['paths']  # different in data model

            for pn, exp_value in partlink_input_props.items():
                assert partlink.properties[pn] == exp_value, (
                    f"Unexpected value for property {pn!r} of partition link:\n"
                    f"{partlink.properties!r}")
            partlink.pull_full_properties()
            for pn, exp_value in partlink_input_props.items():
                assert partlink.properties[pn] == exp_value, (
                    f"Unexpected value for property {pn!r} of partition link:\n"
                    f"{partlink.properties!r}")
            for pn, exp_value in partlink_auto_props.items():
                assert pn in partlink.properties, (
                    f"Automatically returned property {pn!r} is not in "
                    f"created partition link:\n{partlink!r}")
                assert partlink.properties[pn] == exp_value, (
                    f"Unexpected value for property {pn!r} of partition link:\n"
                    f"{partlink.properties!r}")

            # Test updating a property of the partition link

            new_desc = "Updated partition link description."

            # The code to be tested
            partlink.update_properties(dict(description=new_desc))

            assert partlink.properties['description'] == new_desc
            partlink.pull_full_properties()
            assert partlink.properties['description'] == new_desc

            # Test renaming the partition link

            # The code to be tested
            partlink.update_properties(dict(name=partlink_name_new))

            assert partlink.properties['name'] == partlink_name_new
            partlink.pull_full_properties()
            assert partlink.properties['name'] == partlink_name_new
            with pytest.raises(zhmcclient.NotFound):
                console.partition_links.find(name=partlink_name)

        finally:
            # Test deleting the partition link
            if partlink:
                partlink.wait_for_states()

                # The code to be tested
                partlink.delete()

                with pytest.raises(zhmcclient.NotFound):
                    console.partition_links.find(name=partlink_name_new)


PARTLINK_CREATE_DELETE_TESTCASES = [
    # testcases for the test_partlink_create() test function.
    # Each list item is a testcase that is defined with the following
    # tuple items:
    # - desc (string): description of the testcase.
    # - pl_type (string): Type of the partition link, as for its 'type' prop.
    # - input_props (dict): Input properties for the create() method.
    #   The test function adds these properties automatically:
    #   - cpc-uri
    #   - name
    #   - type
    # - exp_props (dict): Expected properties in the created partition link.
    #   The test function adds these properties automatically:
    #   - cpc-uri
    #   - name
    #   - type
    # - exp_exc_type (class): Exception that is expected to be raised during
    #   PartitionLink.create(), or None.
    (
        "Hipersockets with minimal input properties",
        'hipersockets',
        {},
        {
            'bus-connections': [],
            'starting-device-number': '7400',  # default
            'maximum-transmission-unit-size': 8,  # default
        },
        None
    ),
    (
        "SMC-D with minimal input properties",
        'smc-d',
        {},
        {
            'bus-connections': [],
            'starting-fid': 4096,  # default
        },
        None
    ),
    (
        "CTC with minimal input properties",
        'ctc',
        {
            'partitions': ['partition_1.uri', 'partition_2.uri'],
            'paths': [
                {
                    'adapter-port-uri': 'adapter_1.uri',  # Replaced
                    'connecting-adapter-port-uri': 'adapter_1.uri',  # Replaced
                },
            ],
        },
        {
            'paths': [
                {
                    'starting-device-number': '4000',
                    # 'devices': [
                    #     {
                    #         # ctc-endpoint
                    #     }
                    # ],
                    'adapter-port-info': {
                        'adapter-uri': 'adapter_1.uri',  # Replaced
                        'adapter-name': 'adapter_1.name',  # Replaced
                    },
                    'connecting-adapter-port-info': {
                        'adapter-uri': 'adapter_1.uri',  # Replaced
                        'adapter-name': 'adapter_1.name',  # Replaced
                    },
                },
            ],
            'devices-per-path': 4  # default
        },
        None
    ),
]


@pytest.mark.parametrize(
    "desc, pl_type, input_props, exp_props, exp_exc_type",
    PARTLINK_CREATE_DELETE_TESTCASES)
def test_partlink_create_delete(
        dpm_mode_cpcs,  # noqa: F811
        desc, pl_type, input_props, exp_props, exp_exc_type):
    # pylint: disable=redefined-outer-name, exec-used, unused-argument
    """
    Test creation of a partition link (and deletion, for cleanup)
    """
    if not dpm_mode_cpcs:
        pytest.skip("HMC definition does not include any CPCs in DPM mode")

    for cpc in dpm_mode_cpcs:
        assert cpc.dpm_enabled
        skipif_no_partition_link_feature(cpc)

        print(f"Testing on CPC {cpc.name}")

        console = cpc.manager.client.consoles.console
        partlink_name = f"{TEST_PREFIX}_cre_del_{uuid.uuid4().hex}"
        replacements = {}  # Replacements for testcase expressions
        part1 = None
        part2 = None

        # For CTC, get two FC adapters and two stopped partitions
        if pl_type == 'ctc':

            adapters = cpc.adapters.list(
                filter_args={'type': 'fc', 'state': 'online'})
            if len(adapters) < 2:
                pytest.skip(f"CPC {cpc.name} has no two online FC adapters "
                            "for CTC partition link creation")
            adapter_1, adapter_2 = random.choices(adapters, k=2)
            replacements['adapter_1'] = adapter_1
            replacements['adapter_2'] = adapter_2

            part1_name = f"{TEST_PREFIX}_{uuid.uuid4().hex}"
            part1_props = standard_partition_props(cpc, part1_name)
            part1 = cpc.partitions.create(part1_props)
            part2_name = f"{TEST_PREFIX}_{uuid.uuid4().hex}"
            part2_props = standard_partition_props(cpc, part2_name)
            part2 = cpc.partitions.create(part2_props)
            replacements['partition_1'] = part1
            replacements['partition_2'] = part2

        # Prepare the input properties for PartitionLink.create()
        partlink_input_props = {
            'cpc-uri': cpc.uri,
            'name': partlink_name,
            'type': pl_type,
        }
        partlink_input_props.update(input_props)

        # Replace the variable names for adapters and partitions in the
        # input properties with real data.
        partlink_input_props = replace_expressions(
            partlink_input_props, replacements)

        partlink = None
        try:

            if exp_exc_type:
                with pytest.raises(exp_exc_type):

                    # The code to be tested
                    partlink = console.partition_links.create(
                        partlink_input_props)

            else:
                try:

                    # The code to be tested
                    partlink = console.partition_links.create(
                        partlink_input_props)

                except zhmcclient.PartitionLinkError as exc:
                    print("PartitionLinkError during PartitionLink.create(): "
                          f"{exc}")
                    print("Input properties of PartitionLink.create():")
                    pprint(partlink_input_props)
                    raise
                except zhmcclient.HTTPError as exc:
                    print(f"HTTPError during PartitionLink.create(): {exc}")
                    print("Input properties of PartitionLink.create():")
                    pprint(partlink_input_props)
                    raise

                partlink.wait_for_states()

                # Prepare the expected properties
                partlink_exp_props = {
                    'cpc-uri': cpc.uri,
                    'name': partlink_name,
                    'type': pl_type,
                }
                partlink_exp_props.update(exp_props)

                partlink_exp_props = replace_expressions(
                    partlink_exp_props, replacements)

                partlink.pull_full_properties()

                assert_properties(partlink.properties, partlink_exp_props)

        finally:
            # Cleanup, but also code to be tested
            if partlink:
                partlink.wait_for_states()
                partlink.delete()
            if part1:
                part1.delete()
            if part2:
                part2.delete()
            with pytest.raises(zhmcclient.NotFound):
                console.partition_links.find(name=partlink_name)


def test_partlink_zzz_cleanup(dpm_mode_cpcs):  # noqa: F811
    # pylint: disable=redefined-outer-name, exec-used, unused-argument
    """
    Cleanup any created partitions and partition links that may have not been
    cleaned up by the other testcase functions.
    """
    if not dpm_mode_cpcs:
        pytest.skip("HMC definition does not include any CPCs in DPM mode")

    for cpc in dpm_mode_cpcs:
        assert cpc.dpm_enabled
        skipif_no_partition_link_feature(cpc)

        print(f"Testing on CPC {cpc.name}")

        console = cpc.manager.client.consoles.console

        name_pattern = fr'{TEST_PREFIX}.*'
        partlinks = console.partition_links.findall(name=name_pattern)
        for partlink in partlinks:

            print("Deleting test partition link from a prior test run: "
                  f"{partlink.name!r} on CPC {cpc.name} "
                  "- please open a zhmcclient issue for that")

            try:
                partlink.delete()
            except zhmcclient.HTTPError as exc:
                print(f"HTTPError during PartitionLink.delete(): {exc}")

        parts = cpc.partitions.findall(name=name_pattern)
        for part in parts:

            print("Deleting test partition from a prior test run: "
                  f"{part.name!r} on CPC {cpc.name} "
                  "- please open a zhmcclient issue for that")

            try:
                part.delete()
            except zhmcclient.HTTPError as exc:
                print(f"HTTPError during Partition.delete(): {exc}")



# Copyright 2025 IBM Corp. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""
End2end tests for MFA server definitions (on CPCs in DPM mode).

These tests do not change any existing MFA server definitions, but create,
modify and delete test MFA server definitions.
"""


import warnings
import pytest
from requests.packages import urllib3

import zhmcclient
# pylint: disable=line-too-long,unused-import
from zhmcclient.testutils import hmc_definition, hmc_session  # noqa: F401, E501
# pylint: enable=line-too-long,unused-import

from .utils import skip_warn, pick_test_resources, TEST_PREFIX, \
    runtest_find_list, runtest_get_properties

urllib3.disable_warnings()

# Properties in minimalistic MFAServerDefinition objects (e.g. find_by_name())
MFASRVDEF_MINIMAL_PROPS = ['element-uri', 'name']

# Properties in MFAServerDefinition objects returned by list() without full
# props
MFASRVDEF_LIST_PROPS = ['element-uri', 'name']

# Properties whose values can change between retrievals of MFAServerDefinition
# objects
MFASRVDEF_VOLATILE_PROPS = []


def test_mfasrvdef_find_list(hmc_session):  # noqa: F811
    # pylint: disable=redefined-outer-name
    """
    Test list(), find(), findall().
    """
    client = zhmcclient.Client(hmc_session)
    console = client.consoles.console
    hd = hmc_session.hmc_definition

    api_version = client.query_api_version()
    hmc_version = api_version['hmc-version']
    hmc_version_info = tuple(map(int, hmc_version.split('.')))
    if hmc_version_info < (2, 15, 0):
        skip_warn(f"HMC {hd.host} of version {hmc_version} does not yet "
                  "support MFA server definitions")

    # Pick the MFA server definitions to test with
    mfasrvdef_list = console.mfa_server_definitions.list()
    if not mfasrvdef_list:
        skip_warn(f"No MFA server definitions defined on HMC {hd.host}")
    mfasrvdef_list = pick_test_resources(mfasrvdef_list)

    for mfasrvdef in mfasrvdef_list:
        print(f"Testing with MFA server definition {mfasrvdef.name!r}")
        runtest_find_list(
            hmc_session, console.mfa_server_definitions, mfasrvdef.name,
            'name', 'element-uri', MFASRVDEF_VOLATILE_PROPS,
            MFASRVDEF_MINIMAL_PROPS, MFASRVDEF_LIST_PROPS)


def test_mfasrvdef_property(hmc_session):  # noqa: F811
    # pylint: disable=redefined-outer-name
    """
    Test property related methods
    """
    client = zhmcclient.Client(hmc_session)
    console = client.consoles.console
    hd = hmc_session.hmc_definition

    api_version = client.query_api_version()
    hmc_version = api_version['hmc-version']
    hmc_version_info = tuple(map(int, hmc_version.split('.')))
    if hmc_version_info < (2, 15, 0):
        skip_warn(f"HMC {hd.host} of version {hmc_version} does not yet "
                  "support MFA server definitions")

    # Pick the MFA server definitions to test with
    mfasrvdef_list = console.mfa_server_definitions.list()
    if not mfasrvdef_list:
        skip_warn(f"No MFA server definitions defined on HMC {hd.host}")
    mfasrvdef_list = pick_test_resources(mfasrvdef_list)

    for mfasrvdef in mfasrvdef_list:
        print(f"Testing with MFA server definition {mfasrvdef.name!r}")

        # Select a property that is not returned by list()
        non_list_prop = 'description'

        runtest_get_properties(mfasrvdef.manager, non_list_prop)


def test_mfasrvdef_crud(hmc_session):  # noqa: F811
    # pylint: disable=redefined-outer-name
    """
    Test create, read, update and delete a MFA server definition.
    """
    client = zhmcclient.Client(hmc_session)
    console = client.consoles.console
    hd = hmc_session.hmc_definition

    api_version = client.query_api_version()
    hmc_version = api_version['hmc-version']
    hmc_version_info = tuple(map(int, hmc_version.split('.')))
    if hmc_version_info < (2, 15, 0):
        skip_warn(f"HMC {hd.host} of version {hmc_version} does not yet "
                  "support MFA server definitions")

    mfasrvdef_name = TEST_PREFIX + ' test_mfasrvdef_crud mfasrvdef1'
    mfasrvdef_name_new = mfasrvdef_name + ' new'

    # Ensure a clean starting point for this test
    try:
        mfasrvdef = console.mfa_server_definitions.find(
            name=mfasrvdef_name)
    except zhmcclient.NotFound:
        pass
    else:
        warnings.warn(
            "Deleting test MFA server definition from previous run: "
            f"{mfasrvdef_name!r}", UserWarning)
        mfasrvdef.delete()

    mfasrvdef = None
    try:

        # Test creating the MFA server definition

        mfasrvdef_input_props = {
            'name': mfasrvdef_name,
            'description': 'Test MFA server def for zhmcclient end2end tests',
            'hostname-ipaddr': '10.11.12.13',
        }
        mfasrvdef_auto_props = {
            'port': 6789,
            'replication-overwrite-possible': False,
        }

        # The code to be tested
        try:
            mfasrvdef = console.mfa_server_definitions.create(
                mfasrvdef_input_props)
        except zhmcclient.HTTPError as exc:
            if exc.http_status == 403 and exc.reason == 1:
                skip_warn(f"HMC userid {hd.userid!r} is not authorized for "
                          "task 'Manage Multi-factor Authentication' on HMC "
                          f"{hd.host}")
            else:
                raise

        for pn, exp_value in mfasrvdef_input_props.items():
            assert mfasrvdef.properties[pn] == exp_value, \
                f"Unexpected value for property {pn!r}"
        mfasrvdef.pull_full_properties()
        for pn, exp_value in mfasrvdef_input_props.items():
            assert mfasrvdef.properties[pn] == exp_value, \
                f"Unexpected value for property {pn!r}"
        for pn, exp_value in mfasrvdef_auto_props.items():
            assert mfasrvdef.properties[pn] == exp_value, \
                f"Unexpected value for property {pn!r}"

        # Test updating a property of the MFA server definition

        new_desc = "Updated MFA server definition description."

        # The code to be tested
        mfasrvdef.update_properties(dict(description=new_desc))

        assert mfasrvdef.properties['description'] == new_desc
        mfasrvdef.pull_full_properties()
        assert mfasrvdef.properties['description'] == new_desc

        # Test that MFA server definitions can be renamed

        # The code to be tested
        mfasrvdef.update_properties(dict(name=mfasrvdef_name_new))

        with pytest.raises(zhmcclient.NotFound):
            console.mfa_server_definitions.find(name=mfasrvdef_name)

    finally:
        if mfasrvdef:
            # Cleanup and test deleting the MFA server definition

            # The code to be tested
            mfasrvdef.delete()

            with pytest.raises(zhmcclient.NotFound):
                console.mfa_server_definitions.find(name=mfasrvdef_name)

            with pytest.raises(zhmcclient.NotFound):
                console.mfa_server_definitions.find(name=mfasrvdef_name_new)



# Copyright 2025 IBM Corp. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""
End2end tests for hw_messages.

These tests do not delete any existing hardware messages.
"""


import random
import pytest
from requests.packages import urllib3

import zhmcclient
# pylint: disable=line-too-long,unused-import
from zhmcclient.testutils import hmc_definition, hmc_session  # noqa: F401, E501
# pylint: enable=line-too-long,unused-import

from .utils import runtest_find_list, skip_warn, pick_test_resources

urllib3.disable_warnings()

# Properties in minimalistic objects (e.g. find_by_name())
HW_MESSAGE_MINIMAL_PROPS = ['element-uri']

# Properties in objects returned by list() without full props
# Note: 'element-id' is added by the list() method.
HW_MESSAGE_LIST_PROPS = ['element-uri', 'element-id', 'text', 'timestamp']

# Properties whose values can change between retrievals of objects
HW_MESSAGE_VOLATILE_PROPS = []


@pytest.mark.parametrize(
    "parent", ['cpc', 'console']
)
def test_hw_message_find_list(hmc_session, parent):  # noqa: F811
    # pylint: disable=redefined-outer-name
    """
    Test list(), find(), findall().
    """
    client = zhmcclient.Client(hmc_session)
    console = client.consoles.console
    hd = hmc_session.hmc_definition

    if parent == 'cpc':
        parent_obj = random.choice(client.cpcs.list())
        parent_str = f"CPC {parent_obj.name!r}"
    else:
        assert parent == 'console'
        parent_obj = console
        parent_str = "Console"

    # Compare items returned by list() with/without full_properties
    hw_message_list = parent_obj.hw_messages.list()
    hw_message_list_full = parent_obj.hw_messages.list(full_properties=True)
    assert len(hw_message_list) == len(hw_message_list_full)
    hw_message_full_by_id = {}
    for hw_message in hw_message_list_full:
        element_id = hw_message.properties['element-id']
        hw_message_full_by_id[element_id] = hw_message
    for hw_message in hw_message_list:
        element_id = hw_message.properties['element-id']
        assert element_id in hw_message_full_by_id
        hw_message_full = hw_message_full_by_id[element_id]
        for prop_name in HW_MESSAGE_LIST_PROPS:
            assert prop_name in hw_message.properties, (
                f"Property {prop_name!r} not found in hardware-message "
                f"object {element_id!r} returned by list()")
            assert prop_name in hw_message_full.properties, (
                f"Property {prop_name!r} not found in hardware-message "
                f"object {element_id!r} returned by list(full_properties=True)")

    # Pick the hw_messages to test with
    if not hw_message_list:
        skip_warn(f"No hardware messages for {parent_str} on HMC {hd.host}")
    hw_message_list = pick_test_resources(hw_message_list)

    for hw_message in hw_message_list:
        element_id = hw_message.properties['element-id']
        print(f"Testing with hardware message {element_id!r} on {parent_str}")
        runtest_find_list(
            hmc_session, parent_obj.hw_messages, element_id, 'element-id',
            'timestamp', HW_MESSAGE_VOLATILE_PROPS, HW_MESSAGE_MINIMAL_PROPS,
            HW_MESSAGE_LIST_PROPS)


@pytest.mark.parametrize(
    "parent", ['cpc', 'console']
)
def test_hw_message_list_filtered(hmc_session, parent):  # noqa: F811
    # pylint: disable=redefined-outer-name
    """
    Test list() with timestamp filtering.
    """
    client = zhmcclient.Client(hmc_session)
    console = client.consoles.console

    if parent == 'cpc':
        parent_obj = random.choice(client.cpcs.list())
    else:
        assert parent == 'console'
        parent_obj = console

    # Test timestamp filtering
    all_message_list = parent_obj.hw_messages.list()
    sorted_message_list = sorted(all_message_list,
                                 key=lambda m: m.prop('timestamp'))
    begin_index = -3
    end_index = -1
    exp_number = end_index - begin_index + 1
    min_number = exp_number + 2
    if len(sorted_message_list) < min_number:
        skip_warn(f"Not enough hardware messages ({len(sorted_message_list)}) "
                  "for timestamp filtering for {parent_str} on HMC {hd.host}")
    begin_msg = sorted_message_list[begin_index]
    end_msg = sorted_message_list[end_index]
    begin_time = zhmcclient.datetime_from_timestamp(begin_msg.prop('timestamp'))
    end_time = zhmcclient.datetime_from_timestamp(end_msg.prop('timestamp'))

    # The code to be tested.
    filtered_message_list = parent_obj.hw_messages.list(
        begin_time=begin_time, end_time=end_time)

    assert len(filtered_message_list) == exp_number, (
        f"Unexpected number of filtered messages. Begin time: {begin_time}, "
        f"End time: {end_time}, Expected number: {exp_number}, \n"
        f"Actual number: {len(filtered_message_list)!r}"
    )
    for msg in filtered_message_list:
        time = zhmcclient.datetime_from_timestamp(msg.prop('timestamp'))
        assert time >= begin_time
        assert time <= end_time



# Copyright 2025 IBM Corp. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""
Unit tests for _hw_message module.
"""


import re
from datetime import datetime
import pytest

from zhmcclient import Client, NotFound, HwMessage, timestamp_from_datetime
from zhmcclient_mock import FakedSession
from tests.common.utils import assert_resources


class TestHwMessageConsole:
    """
    All tests for the HwMessage and HwMessageManager classes for the Console.
    """

    def setup_method(self):
        """
        Setup that is called by pytest before each test method.

        Set up a faked session, and add a faked Console without any
        child resources.
        """
        # pylint: disable=attribute-defined-outside-init

        self.session = FakedSession('fake-host', 'fake-hmc', '2.13.1', '1.8')
        self.client = Client(self.session)

        self.faked_console = self.session.hmc.consoles.add({
            'element-id': None,
            # element-uri will be automatically set
            'parent': None,
            'class': 'console',
            'name': 'fake-console1',
            'description': 'Console #1',
        })
        self.console = self.client.consoles.find(name=self.faked_console.name)

    def add_hw_message(self, element_id, text, timestamp_dt=None):
        """
        Add a faked hw_message object to the faked Console and return it.
        """
        if timestamp_dt is None:
            timestamp_dt = datetime.now()

        faked_hw_message = self.faked_console.hw_messages.add({
            'element-id': element_id,
            # element-uri will be automatically set
            'parent': None,
            'class': 'hardware-message',
            'text': text,
            'timestamp': timestamp_from_datetime(timestamp_dt),
            'service-supported': True,
            # 'details': ...
        })
        return faked_hw_message

    def test_hw_message_manager_repr(self):
        """Test HwMessageManager.__repr__()."""

        hw_message_mgr = self.console.hw_messages

        # Execute the code to be tested
        repr_str = repr(hw_message_mgr)

        repr_str = repr_str.replace('\n', '\\n')
        # We check just the begin of the string:
        assert re.match(
            rf'^{hw_message_mgr.__class__.__name__}\s+at\s+'
            rf'0x{id(hw_message_mgr):08x}\s+\(\\n.*',
            repr_str)

    def test_hw_message_manager_initial_attrs(self):
        """Test initial attributes of HwMessageManager."""

        hw_message_mgr = self.console.hw_messages

        # Verify all public properties of the manager object
        assert hw_message_mgr.resource_class == HwMessage
        assert hw_message_mgr.class_name == 'hardware-message'
        assert hw_message_mgr.session is self.session
        assert hw_message_mgr.parent is self.console

    @pytest.mark.parametrize(
        "full_properties_kwargs, exp_prop_names", [
            (dict(full_properties=False),
             ['element-uri', 'element-id', 'text', 'timestamp']),
            (dict(full_properties=True),
             ['element-uri', 'element-id', 'text', 'timestamp',
              'service-supported']),
            ({},  # test default for full_properties (False)
             ['element-uri', 'element-id', 'text', 'timestamp']),
        ]
    )
    @pytest.mark.parametrize(
        "filter_args, exp_element_ids", [
            (None,
             ['1', '2']),
            ({},
             ['1', '2']),
            ({'element-id': '1'},
             ['1']),
            ({'element-uri': '/api/console/hardware-messages/1'},
             ['1']),
        ]
    )
    def test_hw_message_manager_list(
            self, filter_args, exp_element_ids, full_properties_kwargs,
            exp_prop_names):
        """Test HwMessageManager.list()."""

        faked_hw_message1 = self.add_hw_message(element_id='1', text='foo')
        faked_hw_message2 = self.add_hw_message(element_id='2', text='bar')
        faked_hw_messages = [faked_hw_message1, faked_hw_message2]
        exp_faked_hw_messages = [
            m for m in faked_hw_messages
            if m.properties['element-id'] in exp_element_ids]
        hw_message_mgr = self.console.hw_messages

        # Execute the code to be tested
        hw_messages = hw_message_mgr.list(filter_args=filter_args,
                                          **full_properties_kwargs)

        assert_resources(hw_messages, exp_faked_hw_messages, exp_prop_names)

    def test_hw_message_repr(self):
        """Test HwMessage.__repr__()."""

        faked_hw_message1 = self.add_hw_message(element_id='1', text='foo')
        hw_message_mgr = self.console.hw_messages
        hw_message1 = hw_message_mgr.find(
            **{'element-id': faked_hw_message1.properties['element-id']})

        # Execute the code to be tested
        repr_str = repr(hw_message1)

        repr_str = repr_str.replace('\n', '\\n')
        # We check just the begin of the string:
        assert re.match(
            rf'^{hw_message1.__class__.__name__}\s+at\s+'
            rf'0x{id(hw_message1):08x}\s+\(\\n.*',
            repr_str)

    def test_hw_message_delete(self):
        """Test HwMessage.delete()."""

        faked_hw_message1 = self.add_hw_message(element_id='1', text='foo')
        self.add_hw_message(element_id='2', text='bar')

        hw_message_mgr = self.console.hw_messages
        hw_message1 = hw_message_mgr.find(
            **{'element-id': faked_hw_message1.properties['element-id']})

        # Execute the code to be tested.
        hw_message1.delete()

        # Check that the HwMessage no longer exists
        with pytest.raises(NotFound):
            hw_message1 = hw_message_mgr.find(
                **{'element-id': faked_hw_message1.properties['element-id']})
